"""Cosmic Ray distributor that sends work requests over HTTP to workers.

This uses a list of URLs to workes, distributing work to them as they're available.

Enabling the distributor
========================

To use the local distributor, set ``cosmic-ray.distributor.name = "http"`` in your Cosmic Ray configuration, and
configure the list of worker URLs in ``cosmic-ray.distributor.http.worker-urls``:

.. code-block:: toml

    [cosmic-ray.distributor]
    name = "http"

    [cosmic-ray.distributor.http]
    worker-urls = ['http://localhost:9876', 'http://localhost:9877']
"""
import asyncio
import logging
from pathlib import Path

import aiohttp
from aiohttp import web
from cosmic_ray.distribution.distributor import Distributor
from cosmic_ray.mutating import mutate_and_test
from cosmic_ray.work_item import MutationSpec, WorkerOutcome, WorkItem, WorkResult

log = logging.getLogger(__name__)


class HttpDistributor(Distributor):
    """The http distributor.

    This forwards mutate-and-test requests to HTTP servers which do the actual work and
    return the results.
    """

    def __call__(self, *args, **kwargs):
        # TODO: We still have the issue that `pending_work` is a running query on the database. Will `on_task_complete`
        # - which writes results to the database - be able to complete, or will it be blocked? Do we have to copy the
        # pending work as we used to do?

        # TODO: Will there be an event loop? Where should we ensure that there is?
        asyncio.get_event_loop().run_until_complete(self._process(*args, **kwargs))

    async def _process(self, pending_work, test_command, timeout, config, on_task_complete):
        urls = config.get("worker-urls", [])

        if not urls:
            raise ValueError("No worker URLs provided for HttpDistributor")

        fetchers = {}

        async def handle_completed_task(task):
            # TODO: If one of the URLs we've got is bad (i.e. no worker is running on it), that will result in an
            # exception from one of the tasks. We should notice this, log it, and remove the offending URL from the
            # pool.

            url, completed_job_id = fetchers[task]
            try:
                result = await task
            except Exception as exc:
                # TODO: Do something with the exception
                log.exception("Error fetching result")
                result = WorkResult(worker_outcome=WorkerOutcome.ABNORMAL, output=str(exc))
            finally:
                del fetchers[task]
                urls.append(url)
                on_task_complete(completed_job_id, result)

        for work_item in pending_work:
            # Wait for an available URL
            while not urls:
                assert fetchers.keys()
                done, pending = await asyncio.wait(fetchers.keys(), return_when=asyncio.FIRST_COMPLETED)
                for task in done:
                    await handle_completed_task(task)

            assert urls, "URL should always be available"

            # Use an available URL to process the task
            url = urls.pop()
            fetcher = asyncio.create_task(send_request(url, work_item, test_command, timeout))
            fetchers[fetcher] = url, work_item.job_id

        # Drain the remaining work
        if fetchers.keys():
            done, pending = await asyncio.wait(fetchers.keys(), return_when=asyncio.ALL_COMPLETED)
            for task in done:
                await handle_completed_task(task)


async def send_request(url, work_item: WorkItem, test_command, timeout):
    """Sends a mutate-and-test request to a worker.

    Args:
        url: The URL of the worker.
        work_item: The `WorkItem` representing the work to be done.
        test_command: The command that the worker should use to run the tests.
        timeout: The maximum number of seconds to spend running the test.

    Returns: A `WorkResult`.
    """
    parameters = {
        "mutations": [
            {
                "module_path": str(mutation.module_path),
                "operator": mutation.operator_name,
                "occurrence": mutation.occurrence,
            }
            for mutation in work_item.mutations
        ],
        "test_command": test_command,
        "timeout": timeout,
    }
    log.info("Sending HTTP request to %s", url)
    async with aiohttp.request("POST", url, json=parameters) as resp:
        result = await resp.json()
        # TODO: Account for possibility that `data` is the wrong shape.
        return WorkResult(
            worker_outcome=result["worker_outcome"],
            output=result["output"],
            test_outcome=result["test_outcome"],
            diff=result["diff"],
        )


async def handle_mutate_and_test(request):
    """HTTP endpoint handler for requests to mutate-and-test."""
    args = await request.json()
    result = await mutate_and_test(
        mutations=[
            MutationSpec(
                module_path=Path(mutation["module_path"]),
                operator_name=mutation["operator"],
                occurrence=mutation["occurrence"],
            )
            for mutation in args["mutations"]
        ],
        test_command=args["test_command"],
        timeout=args["timeout"],
    )
    # TODO: Deal with exceptions. There generally won't be any, so we can just return an abnormal result if there it.

    return web.json_response(
        {
            "worker_outcome": result.worker_outcome.value,
            "output": result.output,
            "test_outcome": result.test_outcome.value if result.test_outcome is not None else None,
            "diff": result.diff,
        }
    )


def run_worker(port=None, path=None):
    """Run the worker HTTP server.

    You must specify either `port` or `path`, but not both.

    Args:
        port: The TCP port on which to listen.
        path: Path to Unix domain socket on which to listen.
    """
    if port is None and path is None:
        raise ValueError("Worker requires either a port or domain socket path")
    app = web.Application()
    app.add_routes([web.post("/", handle_mutate_and_test)])
    web.run_app(app, port=port, path=path)



"Implementation of the zero-iteration-loop operator."

import parso
from parso.python.tree import ForStmt

from .operator import Operator


class ZeroIterationForLoop(Operator):
    """An operator that modified for-loops to have zero iterations."""

    def mutation_positions(self, node):
        if isinstance(node, ForStmt):
            expr = node.children[3]
            yield (expr.start_pos, expr.end_pos)

    def mutate(self, node, index):
        "Modify the For loop to evaluate to None"
        assert index == 0
        assert isinstance(node, ForStmt)

        empty_list = parso.parse(" []")
        node.children[3] = empty_list
        return node

    @classmethod
    def examples(cls):
        return (("for i in rang(1,2): pass", "for i in []: pass"),)



"Tools for querying ASTs."


class ASTQuery:
    """
    Allowing to navigate into any object and test attribute of any object:

    Examples:
    >>> ASTQuery(node).parent.match(Node, type='node').ok

    Test if node.parent isinstance of Node and node.parent.type == 'node'
    At each step (each '.' (dot)) you receive an ObjTest object, then

    Navigation:
    You can call any properties or functions of the base object
    >>> ASTQuery(node).parent.children[2].get_next_sibling()

    Test:
    >>> ASTQuery(node).match(attr='value').match(Class)
    All in once:
    >>> ASTQuery(node).match(Class, attr='value')

    Conditional navigation:
    >>> ASTQuery(node).IF.match(attr='intermediate').parent.FI

    Final result:
    >>> ASTQuery(node).ok
    >>> bool(ASTQuery(node))

    """
    def __init__(self, obj):
        self.obj = obj

    def _clone(self, obj) -> 'ASTQuery':
        "Clone this query."
        return type(self)(obj)

    def match(self, cls=None, **kwargs) -> 'ASTQuery':
        "Check if node matches a class."
        obj = self.obj
        if obj is None:
            return self

        if cls is None or isinstance(obj, cls):
            for k, v in kwargs.items():
                op = None
                k__op = k.split('__')
                if len(k__op) == 2:
                    k, op = k__op
                node_value = getattr(obj, k)
                if op is None:
                    if node_value != v:
                        break
                elif op == 'in':
                    if node_value not in v:
                        break
                else:
                    raise ValueError("Can't handle operator {}".format(op))
            else:
                # All is true, continue recursion
                return self

        # A test fails
        return self._clone(None)

    @property
    def ok(self):
        "Is the query ok."
        return bool(self.obj)

    def __bool__(self):
        return self.ok

    def __getattr__(self, item) -> 'ASTQuery':
        obj = self.obj
        if obj is None:
            return self
        return self._clone(getattr(obj, item))

    @property
    def IF(self):
        "Conditional navigation."
        return ASTQueryOptional(self.obj, obj_test=self)

    def __call__(self, *args, **kwargs) -> 'ASTQuery':
        if self.obj is None:
            return self
        return self._clone(self.obj(*args, **kwargs))

    def __getitem__(self, item) -> 'ASTQuery':
        if self.obj is None:
            return self
        return self._clone(self.obj[item])


class ASTQueryOptional(ASTQuery):
    "Manages conditional navigation."
    def __init__(self, obj, obj_test=None):
        super().__init__(obj)
        self._initial = obj_test

    def _clone(self, obj):
        o = super()._clone(obj)
        o._initial = self._initial  # pylint: disable=protected-access

        return o

    @property
    def FI(self):
        "End of conditional navigation."
        if self:
            return self._initial._clone(self.obj)  # pylint: disable=protected-access
        return self._initial



"A tool for generating HTML reports."

import datetime

from itertools import chain

import click
from yattag import Doc

from cosmic_ray.work_db import WorkDB, use_db
from cosmic_ray.work_item import TestOutcome
from cosmic_ray.tools.survival_rate import kills_count, survival_rate


@click.command()
@click.option("--only-completed/--not-only-completed", default=False)
@click.option("--skip-success/--include-success", default=False)
@click.argument("session-file", type=click.Path(dir_okay=False, readable=True, exists=True))
def report_html(only_completed, skip_success, session_file):
    """Print an HTML formatted report of test results."""
    with use_db(session_file, WorkDB.Mode.open) as db:
        doc = _generate_html_report(db, only_completed, skip_success)

    print(doc.getvalue())


# TODO: Redo this with jinja?


def _generate_html_report(db, only_completed, skip_success):
    # pylint: disable=too-many-statements
    doc, tag, text = Doc().tagtext()
    doc.asis("<!DOCTYPE html>")
    with tag("html", lang="en"):
        with tag("head"):
            doc.stag("meta", charset="utf-8")
            doc.stag("meta", name="viewport", content="width=device-width, initial-scale=1, shrink-to-fit=no")
            doc.stag(
                "link",
                rel="stylesheet",
                href="https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/css/bootstrap.min.css",
                integrity="sha384-ggOyR0iXCbMQv3Xipma34MD+dH/1fQ784/j6cY/iJTQUOhcWr7x9JvoRxT2MZw1T",
                crossorigin="anonymous",
            )
            with tag("title"):
                text("Cosmic Ray Report")
        with tag("body"):
            with tag("div", klass="container"):
                with tag("h1"):
                    with tag("p", klass="text-dark"):
                        text("Cosmic Ray Report")

            all_items = db.completed_work_items
            if not only_completed:
                incomplete = ((item, None) for item in db.pending_work_items)
                all_items = chain(all_items, incomplete)

            with tag("div", klass="container"):

                # Summary info
                _generate_summary(doc, db)

                # Job list

                _generate_job_list(doc, db, skip_success)

            with tag("script"):
                doc.attr(src="https://code.jquery.com/jquery-3.3.1.slim.min.js")
                doc.attr(("integrity", "sha384-q8i/X+965DzO0rT7abK41JStQIAqVgRVzpbzo5smXKp4YfRvH+8abtTE1Pi6jizo"))
                doc.attr(("crossorigin", "anonymous"))
            with tag("script"):
                doc.attr(src="https://cdnjs.cloudflare.com/ajax/libs/popper.js/1.14.7/umd/popper.min.js")
                doc.attr(("integrity", "sha384-UO2eT0CpHqdSJQ6hJty5KVphtPhzWj9WO1clHTMGa3JDZwrnQq4sF86dIHNDz0W1"))
                doc.attr(("crossorigin", "anonymous"))
            with tag("script"):
                doc.attr(src="https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/js/bootstrap.min.js")
                doc.attr(("integrity", "sha384-JjSmVgyd0p3pXB1rRibZUAYoIIy6OrQ6VrjIEaFf/nJGzIxFDsf4x0xIM+B07jRM"))
                doc.attr(("crossorigin", "anonymous"))
            with tag("script", type="text/javascript"):
                doc.asis(
                    "$('div.job_list___sub_multi_collapse').on('shown.bs.collapse',"
                    "  function () {"
                    "    correct_behavior_functional_buttons();"
                    "  });"
                    "$('div.job_list___sub_multi_collapse').on('hidden.bs.collapse',"
                    "  function () {"
                    "    correct_behavior_functional_buttons();"
                    "  });"
                    "function correct_behavior_functional_buttons() {"
                    "    var expand = false;"
                    "    var collapse = false;"
                    "    $('a.job_list___sub_multi_heading').each(function(index) {"
                    "      if ($(this).attr('aria-expanded') == 'false') {"
                    "        expand = true;"
                    "        return false;"
                    "      };"
                    "    });"
                    "    $('a.job_list___sub_multi_heading').each(function(index) {"
                    "      if ($(this).attr('aria-expanded') == 'true') {"
                    "        collapse = true;"
                    "        return false;"
                    "      };"
                    "    });"
                    "    if (expand) {"
                    "      $('div#job_item_expand_all').css('display', 'inline-block');"
                    "    } else {"
                    "      $('div#job_item_expand_all').css('display', 'none');"
                    "    };"
                    "    if (collapse) {"
                    "      $('div#job_item_collapse_all').css('display', 'inline-block');"
                    "    } else {"
                    "      $('div#job_item_collapse_all').css('display', 'none');"
                    "    };"
                    "  };"
                    "correct_behavior_functional_buttons();"
                )

    return doc


def _generate_job_list(doc, db, skip_success):
    doc, tag, text = doc.tagtext()
    with tag("div", klass="mb-1", id="job_list___accordion"):
        with tag("div", klass="card"):
            with tag(
                "a",
                ("data-toggle", "collapse"),
                ("data-target", "#job_list___collapse_1"),
                ("aria-expanded", "false"),
                ("aria-controls", "job_list___collapse_1"),
                href="#",
            ):
                with tag("div", klass="card-header", id="job_list___heading_1"):
                    with tag("button", klass="btn btn-outline-dark"):
                        with tag("h4", klass="m-0"):
                            text("Job list")

            with tag(
                "div",
                ("aria-labelledby", "job_list___heading_1"),
                ("data-parent", "#job_list___accordion"),
                klass="collapse",
                id="job_list___collapse_1",
            ):
                with tag("div", klass="card-body"):
                    with tag("div", klass="text-right mb-1"):
                        with tag("div", klass="mx-1", id="job_item_expand_all"):
                            with tag(
                                "a",
                                href="#",
                                onclick="$('div.job_list___sub_multi_collapse').collapse('show');",
                            ):
                                with tag("button", klass="btn btn-outline-dark"):
                                    with tag("span"):
                                        text("Expand All")
                        with tag("div", klass="mx-1", id="job_item_collapse_all"):
                            with tag(
                                "a",
                                href="#",
                                onclick="$('div.job_list___sub_multi_collapse').collapse('hide');",
                            ):
                                with tag("button", klass="btn btn-outline-dark"):
                                    with tag("span"):
                                        text("Collapse All")

                    # Job item
                    all_items = db.completed_work_items
                    for index, (work_item, result) in enumerate(all_items, start=1):
                        _generate_work_item_card(doc, index, work_item, result, skip_success)


def _generate_work_item_card(doc, index, work_item, result, skip_success):
    doc, tag, text = doc.tagtext()
    if result is not None:
        if result.is_killed:
            if result.test_outcome == TestOutcome.INCOMPETENT:
                level = "info"
            else:
                level = "success"
                if skip_success:
                    return
        else:
            level = "danger"

    with tag("div", klass="mb-1", id="job_list___sub_accordion_{}".format(index)):
        with tag("div", klass="card"):
            with tag(
                "a",
                ("data-toggle", "collapse"),
                ("data-target", "#job_list___sub_collapse_{}_1".format(index)),
                ("aria-expanded", "false"),
                ("aria-controls", "job_list___sub_collapse_{}_1".format(index)),
                href="#",
                klass="job_list___sub_multi_heading",
            ):
                with tag(
                    "div",
                    ("role", "alert"),
                    klass="card-header alert-{}".format(level),
                    id="job_list___sub_heading_{}_1".format(index),
                ):
                    with tag("button", klass="btn btn-outline-{}".format(level)):
                        with tag("span", klass="job_id"):
                            text("{} : Job ID {}".format(index, work_item.job_id))

            with tag(
                "div",
                ("aria-labelledby", "job_list___sub_heading_{}_1".format(index)),
                ("data-parent", "#job_list___sub_accordion_{}".format(index)),
                klass="collapse job_list___sub_multi_collapse",
                id="job_list___sub_collapse_{}_1".format(index),
            ):
                with tag("div", klass="card-body"):
                    with tag("div", klass="work-item"):
                        with tag(
                            "div",
                            klass="alert alert-{} test-outcome".format(level),
                            role="alert",
                        ):
                            if result is not None:
                                if not result.is_killed:
                                    with tag("p"):
                                        text("SURVIVED")
                                with tag("p"):
                                    text("worker outcome: {}".format(result.worker_outcome))
                                with tag("p"):
                                    text("test outcome: {}".format(result.test_outcome))
                            else:
                                with tag("p"):
                                    text("No result")

                    for mutation in work_item.mutations:
                        with tag("pre", klass="location"):
                            with tag(
                                "a",
                                href=pycharm_url(str(mutation.module_path), mutation.start_pos[0]),
                                klass="text-secondary",
                            ):
                                with tag("button", klass="btn btn-outline-dark"):
                                    text(
                                        "{}, start pos: {}, end pos: {}".format(
                                            mutation.module_path,
                                            mutation.start_pos,
                                            mutation.end_pos,
                                        )
                                    )

                        with tag("pre"):
                            text("operator: {}, occurrence: {}".format(mutation.operator_name, mutation.occurrence))

                    if result is not None:
                        if result.diff:
                            with tag("div", klass="alert alert-secondary"):
                                with tag("pre", klass="diff"):
                                    text(result.diff)

                        if result.output:
                            with tag("div", klass="alert alert-secondary"):
                                with tag("pre", klass="diff"):
                                    text(result.output)


def _generate_summary(doc, db):
    doc, tag, text = doc.tagtext()
    num_items = db.num_work_items
    num_complete = db.num_results

    with tag("div", klass="mb-1", id="summary_info___accordion"):
        with tag("div", klass="card"):
            with tag(
                "a",
                ("data-toggle", "collapse"),
                ("data-target", "#summary_info___collapse_1"),
                ("aria-expanded", "true"),
                ("aria-controls", "summary_info___collapse_1"),
                href="#",
            ):
                with tag("div", klass="card-header", id="summary_info___heading_1"):
                    with tag("button", klass="btn btn-outline-dark"):
                        with tag("h4", klass="m-0"):
                            text("Summary info")

            with tag(
                "div",
                ("aria-labelledby", "summary_info___heading_1"),
                ("data-parent", "#summary_info___accordion"),
                klass="collapse show",
                id="summary_info___collapse_1",
            ):
                with tag("div", klass="card-body"):
                    with tag("p"):
                        text("Date time: {}".format(datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")))

                    with tag("p"):
                        text("Total jobs: {}".format(num_items))

                    if num_complete > 0:
                        with tag("p"):
                            text("Complete: {} ({:.2f}%)".format(num_complete, num_complete / num_items * 100))
                        with tag("p"):
                            num_killed = kills_count(db)
                            text("Surviving mutants: {} ({:.2f}%)".format(num_complete - num_killed, survival_rate(db)))
                    else:
                        with tag("p"):
                            text("No jobs completed")


def pycharm_url(filename, line_number):
    "Get a URL for opening a file in Pycharm."
    return "pycharm://open?file={}&line={}".format(filename, line_number)



"""Query and retrieve the various plugins in Cosmic Ray.
"""

import logging

from stevedore import driver, ExtensionManager

log = logging.getLogger()  # pylint: disable=invalid-name


def _log_extension_loading_failure(_mgr, extension_point, err):
    # We have to log at the `error` level here as opposed to, say, `info`
    # because logging isn't configure when we reach here. We need this infor to
    # print with the default logging settings.
    log.error('Operator provider load failure: extension-point="%s", err="%s"', extension_point, err)


OPERATOR_PROVIDERS = {
    extension.name: extension.plugin()
    for extension in ExtensionManager(
        "cosmic_ray.operator_providers", on_load_failure_callback=_log_extension_loading_failure
    )
}


def get_operator(name):
    """Get an operator class from a provider plugin.

    Args:
        name: The name of the operator class.

    Returns:
        The operator *class object* (i.e. not an instance).
    """
    sep = name.index("/")
    provider_name = name[:sep]
    operator_name = name[sep + 1 :]

    provider = OPERATOR_PROVIDERS[provider_name]
    return provider[operator_name]


def operator_names():
    """Get all operator names.

    Returns:
        A sequence of operator names.
    """
    return tuple(
        "{}/{}".format(provider_name, operator_name)
        for provider_name, provider in OPERATOR_PROVIDERS.items()
        for operator_name in provider
    )


def get_distributor(name):
    """Get the distributor by name."""
    manager = driver.DriverManager(
        namespace="cosmic_ray.distributors",
        name=name,
        invoke_on_load=True,
        on_load_failure_callback=_log_extension_loading_failure,
    )

    return manager.driver


def distributor_names():
    """Get all distributor plugin names.

    Returns:
        A sequence of distributor names.
    """
    return ExtensionManager(
        "cosmic_ray.distributors",
        on_load_failure_callback=_log_extension_loading_failure,
    ).names()



"Implementation of operator base class."

from abc import ABC, abstractmethod


class Operator(ABC):
    """The mutation operator base class."""

    @abstractmethod
    def mutation_positions(self, node):
        """All positions where this operator can mutate ``node``.

        An operator might be able to mutate a node in multiple ways, and this
        function should produce a position description for each of these
        mutations. Critically, if an operator can make multiple mutations to the
        same position, this should produce a position for each of these
        mutations (i.e. multiple identical positions).

        Args:
            node: The AST node being mutated.

        Returns:
            An iterable of ``((start-line, start-col), (stop-line, stop-col))``
            tuples describing the locations where this operator will mutate ``node``.
        """

    @abstractmethod
    def mutate(self, node, index):
        """Mutate a node in an operator-specific manner.

        Return the new, mutated node. Return ``None`` if the node has
        been deleted. Return ``node`` if there is no mutation at all for
        some reason.
        """

    @classmethod
    @abstractmethod
    def examples(cls):
        """Examples of the mutations that this operator can make.

        This is primarily for testing purposes, but it could also be used for
        docmentation.

        Each example is a tuple of the form `(from-code, to-code, index)`. The
        `index` is optional and will be assumed to be 0 if it's not included.
        The `from-code` is a string containing some Python code prior to
        mutation. The `to-code` is a string desribing the code after mutation.
        `index` indicates the occurrence of the application of the operator to
        the code (i.e. for when an operator can perform multiple mutation to a
        piece of code).

        Returns: An iterable of example tuples.
        """



"""Functions related to finding modules for testing."""

import glob
from pathlib import Path


def find_modules(module_path):
    """Find all modules in the module (possibly package) represented by ``module_path``.

    Args:
        module_path: A pathlib.Path to a Python package or module.

    Returns:
        An iterable of paths Python modules (i.e. \\*py files).
    """
    if module_path.is_file():
        if module_path.suffix == ".py":
            yield module_path
    elif module_path.is_dir():
        pyfiles = glob.glob("{}/**/*.py".format(module_path), recursive=True)
        yield from (Path(pyfile) for pyfile in pyfiles)


def filter_paths(paths, excluded_paths):
    """Filter out path matching one of excluded_paths glob

    Args:
        paths: path to filter.
        excluded_paths: List for glob of modules to exclude.

    Returns:
        An iterable of paths Python modules (i.e. \\*py files).
    """
    excluded = set(Path(f) for excluded_path in excluded_paths for f in glob.glob(excluded_path, recursive=True))
    return set(paths) - excluded



"""Cosmic Ray distributor that runs tests sequentially and locally.

Enabling the distributor
========================

To use the local distributor, set ``cosmic-ray.distributor.name = "local"`` in your Cosmic Ray configuration:

.. code-block:: toml

    [cosmic-ray.distributor]
    name = "local"
"""

import asyncio
import logging

from cosmic_ray.distribution.distributor import Distributor
from cosmic_ray.mutating import mutate_and_test

log = logging.getLogger(__name__)


class LocalDistributor(Distributor):
    "The local distributor."

    def __call__(self, *args, **kwargs):
        asyncio.get_event_loop().run_until_complete(self._process(*args, **kwargs))

    async def _process(self, pending_work, test_command, timeout, _distributor_config, on_task_complete):
        for work_item in pending_work:
            result = await mutate_and_test(
                mutations=work_item.mutations,
                test_command=test_command,
                timeout=timeout,
            )
            on_task_complete(work_item.job_id, result)



"""
These are the core implementations of the various commands in cosmic ray.

Not all commands are represented here, just the ones which seem big enough to
justify a separate module.
"""

from .execute import execute  # NOQA
from .init import init  # NOQA
from .new_config import new_config  # NOQA



"""Management of progress reporting.

The design of this subsystem is such that progress reporting is
decoupled from updating the current progress. This allows for
progress reporting functions which can be invoked from another
context, such as a SIGINFO signal handler.

As such, reporter callables are responsible only for displaying
current progress, and must be capable of retrieving the latest
progress state from elsewhere when invoked. This may be global state
or reached through references that the reporter callable was
constructed with.  Typically the latest progress state will be
updated by the routine whose progress is being monitored.

To report the current progress invoke report_progress().

To manage installation and deinstallation of progress reporting
functions use the reports_progress() decorator for whole-function
contexts, or the progress_reporter() context manager for narrower
contexts.

It is the responsibility of the client to manage any thread-safety
issues. You should assume that progress reporting functions can be
called asynchronously, at any time, from the main thread.

Example::

  _progress_message = ""

  def _update_foo_progress(i, n):
      global _progress_message
      _progress_message = "{i} of {n} complete".format(i=i, n=n)

  def _report_foo_progress(stream):
      print(_progress_message, file=stream)

  @reports_progress(_report_foo_progress)
  def foo(n):
  for i in range(n):
      _update_foo_progress(i, n)

  # ...

  signal.signal(signal.SIGINFO,
                lambda *args: report_progress())

"""

from contextlib import contextmanager
from functools import wraps

# Currently installed zero-argument callables used to report progress.
import sys

_reporters = []  # pylint: disable=invalid-name


def report_progress(stream=None):
    """Report progress from any currently installed reporters.

    Args:
        stream: The text stream (default: sys.stderr) to which
            progress will be reported.
    """
    if stream is None:
        stream = sys.stderr
    for reporter in _reporters:
        reporter(stream)


@contextmanager
def progress_reporter(reporter):
    """A context manager to install and remove a progress reporting function.

    Args:
        reporter: A zero-argument callable to report progress.
            The callable provided should have the means to both
            retrieve and display current progress information.
    """
    install_progress_reporter(reporter)
    yield
    uninstall_progress_reporter(reporter)


def reports_progress(reporter):
    """A decorator factory to mark functions which report progress.

    Args:
        reporter: A zero-argument callable to report progress.
            The callable provided should have the means to both
            retrieve and display current progress information.
    """

    def decorator(func):  # pylint: disable=missing-docstring
        @wraps(func)
        def wrapper(*args, **kwargs):  # pylint: disable=missing-docstring
            with progress_reporter(reporter):
                return func(*args, **kwargs)

        return wrapper

    return decorator


def install_progress_reporter(reporter):
    """Install a progress reporter.

    Where possible prefer to use the progress_reporter() context
    manager or reports_progress() decorator factory.

    Args:
        reporter: A zero-argument callable to report progress.
            The callable provided should have the means to both
            retrieve and display current progress information.
    """
    _reporters.append(reporter)


def uninstall_progress_reporter(reporter):
    """Uninstall a progress reporter.

    Where possible prefer to use the progress_reporter() context
    manager or reports_progress() decorator factory.

    Args:
        reporter: A callable previously installed by
            install_progress_reporter().
    """
    _reporters.remove(reporter)



"""Implementation of the binary-operator-replacement operator.
"""

from enum import Enum
import itertools

import parso

from .operator import Operator
from .util import extend_name


class BinaryOperators(Enum):
    "All binary operators that we mutate."
    Add = '+'
    Sub = '-'
    Mul = '*'
    Div = '/'
    FloorDiv = '//'
    Mod = '%'
    Pow = '**'
    RShift = '>>'
    LShift = '<<'
    BitOr = '|'
    BitAnd = '&'
    BitXor = '^'


def _create_replace_binary_operator(from_op, to_op):
    @extend_name('_{}_{}'.format(from_op.name, to_op.name))
    class ReplaceBinaryOperator(Operator):
        "An operator that replaces binary {} with binary {}.".format(
            from_op.name, to_op.name)

        def mutation_positions(self, node):
            if _is_binary_operator(node):
                if node.value == from_op.value:
                    yield (node.start_pos, node.end_pos)

        def mutate(self, node, index):
            assert _is_binary_operator(node)
            assert index == 0

            node.value = to_op.value
            return node

        @classmethod
        def examples(cls):
            return (
                ('x {} y'.format(from_op.value), 'x {} y'.format(to_op.value)),
            )

    return ReplaceBinaryOperator


# Parent types of operators which indicate that the operator isn't binary.
_NON_BINARY_PARENTS = set((
    'factor',       # unary operators, e.g. -1
    'argument',     # extended function definitions, e.g. def foo(*args)
    'star_expr',    # destructuring, e.g. a, *b = x
    'import_from',  # star import, e.g. from module import *
))


def _is_binary_operator(node):
    if isinstance(node, parso.python.tree.Operator):
        # This catches extended call syntax, e.g. call(*x)
        if isinstance(node.parent, parso.python.tree.Param):
            return False

        if isinstance(node.parent, parso.python.tree.PythonNode):
            return node.parent.type not in _NON_BINARY_PARENTS

        return True

    return False


# Build all of the binary replacement operators
_MUTATION_OPERATORS = tuple(
    _create_replace_binary_operator(from_op, to_op)
    for from_op, to_op
    in itertools.permutations(BinaryOperators, 2))

# Inject operators into module namespace
for op_cls in _MUTATION_OPERATORS:
    globals()[op_cls.__name__] = op_cls


def operators():
    "Iterable of all binary operator replacement mutation operators."
    return _MUTATION_OPERATORS



"""An filter that removes operators based on regular expressions.
"""
from argparse import Namespace
import logging
import re
import sys

from cosmic_ray.config import ConfigDict, load_config
from cosmic_ray.work_db import WorkDB
from cosmic_ray.work_item import WorkerOutcome, WorkResult
from cosmic_ray.tools.filters.filter_app import FilterApp

log = logging.getLogger()


class OperatorsFilter(FilterApp):
    "Implemenents the operators-filter."

    def description(self):
        return __doc__

    def _skip_filtered(self, work_db, exclude_operators):
        if not exclude_operators:
            return

        re_exclude_operators = re.compile("|".join("(:?%s)" % e for e in exclude_operators))

        for item in work_db.pending_work_items:
            for mutation in item.mutations:
                if re_exclude_operators.match(mutation.operator_name):
                    log.info(
                        "operator skipping %s %s %s %s %s %s",
                        item.job_id,
                        mutation.operator_name,
                        mutation.occurrence,
                        mutation.module_path,
                        mutation.start_pos,
                        mutation.end_pos,
                    )

                    work_db.set_result(
                        item.job_id,
                        WorkResult(
                            output="Filtered operator",
                            worker_outcome=WorkerOutcome.SKIPPED,
                        ),
                    )

                    break

    def filter(self, work_db: WorkDB, args: Namespace):
        """Mark as skipped all work item with filtered operator"""

        config = ConfigDict()
        if args.config is not None:
            config = load_config(args.config)

        exclude_operators = config.sub("filters", "operators-filter").get("exclude-operators", ())
        self._skip_filtered(work_db, exclude_operators)

    def add_args(self, parser):
        parser.add_argument("--config", help="Config file to use")


def main(argv=None):
    """Run the operators-filter with the specified command line arguments."""
    return OperatorsFilter().main(argv)


if __name__ == "__main__":
    sys.exit(main())



"""Implementation of the WorkDB."""

import contextlib
from pathlib import Path

from sqlalchemy import Column, Enum, ForeignKey, Integer, String, Text, create_engine, event
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import declarative_base, relationship
from sqlalchemy.orm.session import sessionmaker

from .work_item import ResolvedMutationSpec, TestOutcome, WorkResult, WorkerOutcome, WorkItem


class WorkDB:
    """WorkDB is the database that keeps track of mutation testing work progress.

    Essentially, there's a row in the DB for each mutation that needs to be
    executed in some run. These initially start off with no results, and
    results are added as they're completed.
    """

    class Mode(Enum):
        "Modes in which a WorkDB may be opened."

        # Open existing files, creating if necessary
        create = 1

        # Open only existing files, failing if it doesn't exist
        open = 2

    def __init__(self, path, mode):
        """Open a DB in file `path` in mode `mode`.

        Args:
          path: The path to the DB file.

        Raises:
          FileNotFoundError: If `mode` is `Mode.open` and `path` does not
            exist.
        """
        self._path = path
        if mode == WorkDB.Mode.open and (not Path(path).exists()):
            raise FileNotFoundError("File does not exist: {}".format(path))

        self._engine = create_engine("sqlite:///{}".format(path))

        def enable_foreign_keys(dbapi_con, _con_rec):
            dbapi_con.execute("pragma foreign_keys=ON")

        event.listen(self._engine, "connect", enable_foreign_keys)
        Base.metadata.create_all(self._engine)
        self._session_maker = sessionmaker(self._engine)

    def close(self):
        """Close the database."""
        # TODO: We don't really need this any more. Should we just remove it? Similarly, is there any need
        # for use_db()? It seems possible that we'll eventually move back to needing an explicit close()
        # call, so I'm a bit hesitant to remove this.

    def name(self):
        """A name for this database.

        Derived from the constructor arguments.
        """
        return str(self._path)

    @property
    def work_items(self):
        """An iterable of all of WorkItems in the db.

        This includes both WorkItems with and without results.
        """
        with self._session_maker.begin() as session:
            return tuple(_work_item_from_storage(work_item) for work_item in session.query(WorkItemStorage).all())

    @property
    def num_work_items(self):
        """The number of work items."""
        with self._session_maker.begin() as session:
            return session.query(WorkItemStorage).count()

    def add_work_item(self, work_item):
        """Add a :class:`WorkItem`.

        Args:
          work_item: A ``WorkItem``.
        """
        self.add_work_items((work_item,))

    def add_work_items(self, work_items):
        """Add multiple WorkItems.

        Args:
          work_items: an iterable of WorkItem.
        """
        storage = (_work_item_to_storage(work_item) for work_item in work_items)

        with self._session_maker.begin() as session:
            session.add_all(storage)

    def clear(self):
        """Clear all work items from the session.

        This removes any associated results as well.
        """
        with self._session_maker.begin() as session:
            session.query(WorkResultStorage).delete()
            session.query(MutationSpecStorage).delete()
            session.query(WorkItemStorage).delete()

    @property
    def results(self):
        "An iterable of all ``(job-id, WorkResult)``\\s."
        with self._session_maker.begin() as session:
            for result in session.query(WorkResultStorage).all():
                yield result.job_id, _work_result_from_storage(result)

    @property
    def num_results(self):
        """The number of results."""
        with self._session_maker.begin() as session:
            return session.query(WorkResultStorage).count()

    def set_result(self, job_id, result):
        """Set the result for a job.

        This will overwrite any existing results for the job.

        Args:
          job_id: The ID of the WorkItem to set the result for.
          result: A WorkResult indicating the result of the job.

        Raises:
           KeyError: If there is no work-item with a matching job-id.
        """
        try:
            with self._session_maker.begin() as session:
                storage = _work_result_to_storage(result, job_id)
                session.merge(storage)
        except IntegrityError:
            raise KeyError("Unable to add results for job-id {}. No matching WorkItem.".format(job_id))

    @property
    def pending_work_items(self):
        "Iterable of all pending work items. In random order."
        with self._session_maker.begin() as session:
            completed_job_ids = session.query(WorkResultStorage.job_id)
            pending = session.query(WorkItemStorage).where(~WorkItemStorage.job_id.in_(completed_job_ids))
            return tuple(_work_item_from_storage(work_item) for work_item in pending)

    @property
    def completed_work_items(self):
        "Iterable of ``(work-item, result)``\\s for all completed items."
        with self._session_maker.begin() as session:
            results = session.query(WorkItemStorage, WorkResultStorage).where(
                WorkItemStorage.job_id == WorkResultStorage.job_id
            )
            return tuple(
                (_work_item_from_storage(work_item), _work_result_from_storage(result)) for work_item, result in results
            )


@contextlib.contextmanager
def use_db(path, mode=WorkDB.Mode.create):
    """
    Open a DB in file `path` in mode `mode` as a context manager.

    On exiting the context the DB will be automatically closed.

    Args:
      path: The path to the DB file.

    Raises:
      FileNotFoundError: If `mode` is `Mode.open` and `path` does not
        exist.
    """
    database = WorkDB(path, mode)
    try:
        yield database

    finally:
        database.close()


Base = declarative_base()


class WorkItemStorage(Base):
    "Database model for WorkItem."
    __tablename__ = "work_items"

    job_id = Column(String, primary_key=True)
    mutations = relationship("MutationSpecStorage", back_populates="work_item")


class MutationSpecStorage(Base):
    "Database model for MutationSpecs"
    __tablename__ = "mutation_specs"
    module_path = Column(String)
    operator_name = Column(String)
    occurrence = Column(Integer)
    start_pos_row = Column(Integer)
    start_pos_col = Column(Integer)
    end_pos_row = Column(Integer)
    end_pos_col = Column(Integer)
    job_id = Column(String, ForeignKey("work_items.job_id"), primary_key=True)
    work_item = relationship("WorkItemStorage", back_populates="mutations")


class WorkResultStorage(Base):
    "Database model for WorkResult."
    __tablename__ = "work_results"

    worker_outcome = Column(Enum(WorkerOutcome))
    output = Column(Text, nullable=True)
    test_outcome = Column(Enum(TestOutcome), nullable=True)
    diff = Column(Text, nullable=True)
    job_id = Column(String, ForeignKey("work_items.job_id"), primary_key=True)


def _mutation_spec_from_storage(mutation_spec: MutationSpecStorage):
    return ResolvedMutationSpec(
        module_path=Path(mutation_spec.module_path),
        operator_name=mutation_spec.operator_name,
        occurrence=mutation_spec.occurrence,
        start_pos=(mutation_spec.start_pos_row, mutation_spec.start_pos_col),
        end_pos=(mutation_spec.end_pos_row, mutation_spec.end_pos_col),
    )


def _mutation_spec_to_storage(mutation_spec: ResolvedMutationSpec, job_id: str):
    return MutationSpecStorage(
        job_id=job_id,
        module_path=str(mutation_spec.module_path),
        operator_name=mutation_spec.operator_name,
        occurrence=mutation_spec.occurrence,
        start_pos_row=mutation_spec.start_pos[0],
        start_pos_col=mutation_spec.start_pos[1],
        end_pos_row=mutation_spec.end_pos[0],
        end_pos_col=mutation_spec.end_pos[1],
    )


def _work_item_from_storage(work_item: WorkItemStorage):
    return WorkItem(
        mutations=tuple(_mutation_spec_from_storage(s) for s in work_item.mutations),
        job_id=work_item.job_id,
    )


def _work_item_to_storage(work_item: WorkItem):
    return WorkItemStorage(
        job_id=work_item.job_id,
        mutations=[_mutation_spec_to_storage(mutation, work_item.job_id) for mutation in work_item.mutations],
    )


def _work_result_to_storage(result: WorkResult, job_id):
    return WorkResultStorage(
        worker_outcome=result.worker_outcome,
        output=result.output,
        test_outcome=result.test_outcome,
        diff=result.diff,
        job_id=job_id,
    )


def _work_result_from_storage(result: WorkResultStorage):
    return WorkResult(
        worker_outcome=result.worker_outcome,
        output=result.output,
        test_outcome=result.test_outcome,
        diff=result.diff,
    )



"""Tool for creating badge.
"""
import os
from logging import getLogger

from anybadge import Badge

import click
from cosmic_ray.config import load_config
from cosmic_ray.tools.survival_rate import survival_rate
from cosmic_ray.work_db import WorkDB, use_db

log = getLogger()


@click.command()
@click.argument("config_file", type=click.Path(exists=True, dir_okay=False, readable=True))
@click.argument("badge_file", type=click.Path(dir_okay=False, writable=True))
@click.argument("session_file", type=click.Path(exists=True, dir_okay=False, readable=True))
def generate_badge(config_file, badge_file, session_file):
    """Generate badge file."""

    with use_db(session_file, WorkDB.Mode.open) as db:
        config = load_config(config_file)

        percent = 100 - survival_rate(db)

        config = config["badge"]

        badge = Badge(
            label=config["label"],
            value=percent,
            value_format=config["format"],
            thresholds=config["thresholds"],
        )

        log.info("Generating badge: " + config["format"], percent)  # pylint: disable=logging-not-lazy

        try:
            os.unlink(badge_file)
        except OSError:
            pass

        badge.write_badge(badge_file)



"Implementation of the 'init' command."
import logging
from typing import Iterable
import uuid

from cosmic_ray.ast import get_ast, ast_nodes
import cosmic_ray.modules
from cosmic_ray.work_item import MutationSpec, ResolvedMutationSpec, WorkItem
from cosmic_ray.plugins import get_operator
from cosmic_ray.work_db import WorkDB

log = logging.getLogger()


def _all_work_items(module_paths, operator_names) -> Iterable[WorkItem]:
    "Iterable of all WorkItems for the given inputs."
    for module_path in module_paths:
        module_ast = get_ast(module_path)

        for op_name in operator_names:
            operator = get_operator(op_name)()

            positions = (
                (start_pos, end_pos)
                for node in ast_nodes(module_ast)
                for start_pos, end_pos in operator.mutation_positions(node)
            )

            for occurrence, (start_pos, end_pos) in enumerate(positions):
                mutation = ResolvedMutationSpec(
                    module_path=str(module_path),
                    operator_name=op_name,
                    occurrence=occurrence,
                    start_pos=start_pos,
                    end_pos=end_pos,
                )

                yield WorkItem.single(job_id=uuid.uuid4().hex, mutation=mutation)


def init(module_paths, work_db: WorkDB):
    """Clear and initialize a work-db with work items.

    Any existing data in the work-db will be cleared and replaced with entirely
    new work orders. In particular, this means that any results in the db are
    removed.

    Args:
      module_paths: iterable of pathlib.Paths of modules to mutate.
      work_db: A `WorkDB` instance into which the work orders will be saved.
    """

    operator_names = list(cosmic_ray.plugins.operator_names())

    work_db.clear()
    work_db.add_work_items(_all_work_items(module_paths, operator_names))



"""This is the command-line program for cosmic ray.

Here we manage command-line parsing and launching of the internal machinery that does mutation testing.
"""
import asyncio
import dataclasses
import json
import logging
import os
import signal
import subprocess
import sys
from collections import defaultdict
from contextlib import contextmanager, redirect_stdout
from pathlib import Path
import tempfile

import click
from exit_codes import ExitCode
from rich.logging import RichHandler

import cosmic_ray.commands
import cosmic_ray.modules
import cosmic_ray.mutating
import cosmic_ray.plugins
import cosmic_ray.testing
import cosmic_ray.distribution.http
from cosmic_ray.config import load_config, serialize_config
from cosmic_ray.mutating import apply_mutation
from cosmic_ray.progress import report_progress
from cosmic_ray.version import __version__
from cosmic_ray.work_db import use_db, WorkDB
from cosmic_ray.work_item import TestOutcome, WorkItem

log = logging.getLogger()


@click.group()
@click.option(
    "--verbosity",
    default="WARNING",
    help="The logging level to use.",
    type=click.Choice(["CRITICAL", "DEBUG", "ERROR", "FATAL", "INFO", "WARNING"], case_sensitive=True),
)
@click.version_option(version=__version__)
def cli(verbosity):
    "Mutation testing for Python3"
    logging_level = getattr(logging, verbosity)
    logging.basicConfig(level=logging_level, handlers=[RichHandler()])

    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())


@cli.command()
@click.argument("config_file", type=click.File("wt"))
def new_config(config_file):
    """Create a new config file."""
    cfg = cosmic_ray.commands.new_config()
    config_str = serialize_config(cfg)
    config_file.write(config_str)
    sys.exit(ExitCode.OK)


click.argument()


@cli.command()
@click.argument("config_file")
@click.argument(
    "session_file",
    # help="The filename for the database in which the work order will be stored."
)
def init(config_file, session_file):
    """Initialize a mutation testing session from a configuration. This
    primarily creates a session - a database of "work to be done" -
    which describes all of the mutations and test runs that need to be
    executed for a full mutation testing run. The configuration
    specifies the top-level module to mutate, the tests to run, and how
    to run them.

    This command doesn't actually run any tests. Instead, it scans the
    modules-under-test and simply generates the work order which can be
    executed with other commands.
    """
    cfg = load_config(config_file)

    modules = cosmic_ray.modules.find_modules(Path(cfg["module-path"]))
    modules = cosmic_ray.modules.filter_paths(modules, cfg.get("exclude-modules", ()))

    if log.isEnabledFor(logging.INFO):
        log.info("Modules discovered:")
        per_dir = defaultdict(list)
        for m in modules:
            per_dir[m.parent].append(m.name)
        for directory, files in per_dir.items():
            log.info(" - %s: %s", directory, ", ".join(sorted(files)))

    with use_db(session_file) as database:
        cosmic_ray.commands.init(modules, database)

    sys.exit(ExitCode.OK)


@cli.command(name="exec")
@click.argument("config_file")
@click.argument("session_file")
def handle_exec(config_file, session_file):
    """Perform the remaining work to be done in the specified session.
    This requires that the rest of your mutation testing
    infrastructure (e.g. worker processes) are already running.
    """
    cfg = load_config(config_file)

    with use_db(session_file, mode=WorkDB.Mode.open) as work_db:
        cosmic_ray.commands.execute(work_db, cfg)
    sys.exit(ExitCode.OK)


@cli.command()
@click.argument("config_file")
@click.option(
    "--session-file",
    type=click.Path(dir_okay=False, writable=True, path_type=Path),
    help="Path to session file. If not provided, a temp file is used.",
)
def baseline(config_file, session_file):
    """Runs a baseline execution that executes the test suite over unmutated code.

    If ``--session-file`` is provided, the session used for baselining is stored in that file. Otherwise,
    the session is stored in a temporary file which is deleted after the baselining.

    Exits with 0 if the job has exited normally, otherwise 1.
    """
    cfg = load_config(config_file)

    @contextmanager
    def path_or_temp(path):
        if path is None:
            with tempfile.TemporaryDirectory() as tmpdir:
                yield Path(tmpdir) / "session.sqlite"
        else:
            yield path

    with path_or_temp(session_file) as session_path:
        with use_db(session_path, mode=WorkDB.Mode.create) as db:
            db.clear()
            db.add_work_item(
                WorkItem(
                    mutations=[],
                    job_id="baseline",
                )
            )

            # Run the single-entry session.
            cosmic_ray.commands.execute(db, cfg)

            result = next(db.results)[1]
            if result.test_outcome == TestOutcome.KILLED:
                message = ["Baseline failed. Execution with no mutation gives those following errors:"]
                for line in result.output.split("\n"):
                    message.append("  >>> {}".format(line))
                log.error("\n".join(message))
                sys.exit(1)
            else:
                log.info("Baseline passed. Execution with no mutation works fine.")
                sys.exit(ExitCode.OK)


@cli.command()
@click.argument("session_file")
def dump(session_file):
    """JSON dump of session data. This output is typically run through other
    programs to produce reports.

    Each line of output is a list with two elements: a WorkItem and a
    WorkResult, both JSON-serialized. The WorkResult can be null, indicating a
    WorkItem with no results.
    """

    def item_to_dict(work_item):
        d = dataclasses.asdict(work_item)
        for m in d["mutations"]:
            m["module_path"] = str(m["module_path"])
        return d

    def result_to_dict(result):
        d = dataclasses.asdict(result)
        d["worker_outcome"] = d["worker_outcome"].value
        d["test_outcome"] = d["test_outcome"].value
        return d

    with use_db(session_file, WorkDB.Mode.open) as database:
        for work_item, result in database.completed_work_items:
            print(json.dumps((item_to_dict(work_item), result_to_dict(result))))
        for work_item in database.pending_work_items:
            print(json.dumps((item_to_dict(work_item), None)))

    sys.exit(ExitCode.OK)


@cli.command()
def operators():
    """List the available operator plugins."""
    print("\n".join(cosmic_ray.plugins.operator_names()))

    sys.exit(ExitCode.OK)


@cli.command()
def distributors():
    """List the available distributor plugins."""
    print("\n".join(cosmic_ray.plugins.distributor_names()))

    sys.exit(ExitCode.OK)


@cli.command()
@click.argument("module_path")
@click.argument("operator")
@click.argument("occurrence", type=int)
def apply(module_path, operator, occurrence):
    """Apply the specified mutation to the files on disk. This is primarily a debugging tool."""

    apply_mutation(Path(module_path), cosmic_ray.plugins.get_operator(operator)(), occurrence)

    sys.exit(ExitCode.OK)


@cli.command()
@click.option("--port", type=int, default=None, help="The port on which to listen for requests")
@click.option("--path", default=None, help="Path to Unix domain socket on which to listen for requests")
def http_worker(port, path):
    """Run an HTTP worker for the 'http' distributor."""
    if (port is None) == (path is None):
        log.error("You must specify exactly one of --path or --port")
        sys.exit(ExitCode.USAGE)

    try:
        cosmic_ray.distribution.http.run_worker(port=port, path=path)
    except ValueError as exc:
        log.error(str(exc))
        sys.exit(ExitCode.DATA_ERR)

    sys.exit(ExitCode.OK)


@cli.command()
@click.argument("module_path")
@click.argument("operator")
@click.argument("occurrence", type=int)
@click.argument("test_command")
@click.option("--keep-stdout", default=False, flag_value=True, help="Do not squelch output.")
def mutate_and_test(module_path, operator, occurrence, test_command, keep_stdout):
    """Run a worker process which performs a single mutation and test run.
    Each worker does a minimal, isolated chunk of work: it mutates the
    <occurrence>-th instance of <operator> in <module-path>, runs the test
    suite defined in the configuration, prints the results, and exits.

    Normally you won't run this directly. Rather, it will be launched
    by an distributor. However, it can be useful to run this on
    its own for testing and debugging purposes.
    """
    with open(os.devnull, "w") as devnull:
        with redirect_stdout(sys.stdout if keep_stdout else devnull):
            work_result = asyncio.get_event_loop().run_until_complete(
                cosmic_ray.mutating.mutate_and_test(Path(module_path), operator, occurrence, test_command, None)
            )

    sys.stdout.write(json.dumps(dataclasses.asdict(work_result)))

    sys.exit(ExitCode.OK)


_SIGNAL_EXIT_CODE_BASE = 128


def main(argv=None):
    """Invoke the cosmic ray evaluation.

    :param argv: the command line arguments
    """
    signal.signal(signal.SIGINT, lambda *args: sys.exit(_SIGNAL_EXIT_CODE_BASE + signal.SIGINT))

    if hasattr(signal, "SIGINFO"):
        signal.signal(getattr(signal, "SIGINFO"), lambda *args: report_progress(sys.stderr))

    try:
        return cli(argv)
    except FileNotFoundError as exc:
        print(exc, file=sys.stderr)
        return ExitCode.NO_INPUT
    except PermissionError as exc:
        print(exc, file=sys.stderr)
        return ExitCode.NO_PERM
    except cosmic_ray.config.ConfigError as exc:
        print(repr(exc), file=sys.stderr)
        if exc.__cause__ is not None:
            print(exc.__cause__, file=sys.stderr)
        return ExitCode.CONFIG
    except subprocess.CalledProcessError as exc:
        print("Error in subprocess", file=sys.stderr)
        print(exc, file=sys.stderr)
        return exc.returncode
    except SystemExit as exc:
        # We intercept this here so that main() is testable.
        return exc.code


if __name__ == "__main__":
    sys.exit(main())



"""A filter that uses git to determine when specific files/lines
should be skipped.
"""
import logging
import re
import subprocess
import sys
from argparse import Namespace
from collections import defaultdict
from functools import lru_cache
from pathlib import Path

from cosmic_ray.config import ConfigDict, load_config
from cosmic_ray.work_db import WorkDB
from cosmic_ray.work_item import WorkerOutcome, WorkResult
from cosmic_ray.tools.filters.filter_app import FilterApp

log = logging.getLogger()


class GitFilter(FilterApp):
    """Implements the git filter."""

    def description(self):
        return __doc__

    def _git_news(self, branch):
        """Get the set of new lines by file"""
        # we could use interlap, but do not want to
        # add new dependency at the moment
        diff = subprocess.run(["git", "diff", "--relative", "-U0", branch, "."], capture_output=True)
        regex = re.compile(r"@@ -\d+(?:,\d+)? \+(\d+)(?:,(\d+))? @@.*")
        current_file = None
        res = defaultdict(set)
        for diff_line in diff.stdout.decode("utf-8").split("\n"):
            if diff_line.startswith("@@"):
                m = regex.match(diff_line)
                start = int(m.group(1))
                lenght = int(m.group(2)) if m.group(2) is not None else 1
                for line in range(start, start + lenght):
                    res[current_file].add(line)
            if diff_line.startswith("+++ b/"):
                current_file = Path(diff_line[6:])
        return res

    def _skip_filtered(self, work_db, branch):
        git_news = self._git_news(branch)

        for item in work_db.pending_work_items:
            for mutation in item.mutations:
                if mutation.module_path not in git_news or not (
                    git_news[mutation.module_path] & set(range(mutation.start_pos[0], mutation.end_pos[0] + 1))
                ):
                    log.info(
                        "git skipping %s %s %s %s %s %s",
                        item.job_id,
                        mutation.operator_name,
                        mutation.occurrence,
                        mutation.module_path,
                        mutation.start_pos,
                        mutation.end_pos,
                    )

                    work_db.set_result(
                        item.job_id,
                        WorkResult(
                            output="Filtered git",
                            worker_outcome=WorkerOutcome.SKIPPED,
                        ),
                    )

    def filter(self, work_db: WorkDB, args: Namespace):
        """Mark as skipped all work item that is not new"""

        config = ConfigDict()
        if args.config is not None:
            config = load_config(args.config)

        branch = config.sub("git", "git-filter").get("branch", "master")
        self._skip_filtered(work_db, branch)

    def add_args(self, parser):
        parser.add_argument("--config", help="Config file to use")


def main(argv=None):
    """Run the operators-filter with the specified command line arguments."""
    return GitFilter().main(argv)


if __name__ == "__main__":
    sys.exit(main())


"""Support for making mutations to source code.
"""
import contextlib
import difflib
from itertools import chain
import traceback
from contextlib import contextmanager
import logging
from typing import Iterable

import cosmic_ray.plugins
from cosmic_ray.ast import Visitor, get_ast
from cosmic_ray.testing import run_tests
from cosmic_ray.work_item import MutationSpec, TestOutcome, WorkerOutcome, WorkResult

log = logging.getLogger(__name__)

# pylint: disable=R0913
async def mutate_and_test(mutations: Iterable[MutationSpec], test_command, timeout) -> WorkResult:
    """Apply a sequence of mutations, run thest tests, and reports the results.

    This is fundamentally the mutation(s)-and-test-run implementation at the heart of Cosmic Ray.

    There are three high-level ways that a worker can finish. First, it could fail exceptionally, meaning that some
    uncaught exception made its way from some part of the operation to terminate the function. This function will
    intercept all exceptions and return it in a non-exceptional structure.

    Second, the mutation machinery may determines that - for any of the mutations - there is mutation to be made (e.g.
    the 'occurrence' is too high).  In this case there is no way to report a test result (i.e. killed, survived, or
    incompetent) so a special value is returned indicating that no mutation is possible.

    Finally, and hopefully normally, the worker will find that it can run a test. It will do so and report back the
    result - killed, survived, or incompetent - in a structured way.

    Args:
        mutations: An iterable of ``MutationSpec``\\s describing the mutations to make.
        test_command: The command to execute to run the tests
        timeout: The maximum amount of time (seconds) to let the tests run

    Returns:
        A ``WorkResult``.

    Raises:
        This will generally not raise any exceptions. Rather, exceptions will be reported using the 'exception'
        result-type in the return value.

    """
    try:
        with contextlib.ExitStack() as stack:
            file_changes = {}
            for mutation in mutations:
                operator_class = cosmic_ray.plugins.get_operator(mutation.operator_name)
                operator = operator_class()
                (previous_code, mutated_code) = stack.enter_context(
                    use_mutation(mutation.module_path, operator, mutation.occurrence)
                )

                # If there's no mutated code, then no mutation was possible.
                if mutated_code is None:
                    return WorkResult(
                        worker_outcome=WorkerOutcome.NO_TEST,
                    )

                original_code, _ = file_changes.get(mutation.module_path, (previous_code, mutated_code))
                file_changes[mutation.module_path] = original_code, mutated_code

            test_outcome, output = await run_tests(test_command, timeout)

            diffs = [
                _make_diff(original_code, mutated_code, module_path)
                for module_path, (original_code, mutated_code) in file_changes.items()
            ]

            return WorkResult(
                output=output,
                diff="\n".join(chain(*diffs)),
                test_outcome=test_outcome,
                worker_outcome=WorkerOutcome.NORMAL,
            )

    except Exception:  # noqa # pylint: disable=broad-except
        return WorkResult(
            output=traceback.format_exc(), test_outcome=TestOutcome.INCOMPETENT, worker_outcome=WorkerOutcome.EXCEPTION
        )


@contextmanager
def use_mutation(module_path, operator, occurrence):
    """A context manager that applies a mutation for the duration of a with-block.

    This applies a mutation to a file on disk, and after the with-block it put the unmutated code
    back in place.

    Args:
        module_path: The path to the module to mutate.
        operator: The `Operator` instance to use.
        occurrence: The occurrence of the operator to apply.

    Yields:
        A `(unmutated-code, mutated-code)` tuple to the with-block. If there was no
        mutation performed, the `mutated-code` is `None`.
    """
    # TODO: Could/should use async?
    original_code, mutated_code = apply_mutation(module_path, operator, occurrence)
    try:
        yield original_code, mutated_code
    finally:
        with module_path.open(mode="wt", encoding="utf-8") as handle:
            handle.write(original_code)
            handle.flush()


def apply_mutation(module_path, operator, occurrence):
    """Apply a specific mutation to a file on disk.

    Args:
        module_path: The path to the module to mutate.
        operator: The `operator` instance to use.
        occurrence: The occurrence of the operator to apply.

    Returns:
        A `(unmutated-code, mutated-code)` tuple to the with-block. If there was
        no mutation performed, the `mutated-code` is `None`.
    """
    log.info("Applying mutation: path=%s, op=%s, occurrence=%s", module_path, operator, occurrence)
    module_ast = get_ast(module_path)
    original_code = module_ast.get_code()
    visitor = MutationVisitor(occurrence, operator)
    mutated_ast = visitor.walk(module_ast)

    mutated_code = None
    if visitor.mutation_applied:
        mutated_code = mutated_ast.get_code()
        with module_path.open(mode="wt", encoding="utf-8") as handle:
            handle.write(mutated_code)
            handle.flush()

    return original_code, mutated_code


class MutationVisitor(Visitor):
    """Visitor that mutates a module with the specific occurrence of an operator.

    This will perform at most one mutation in a walk of an AST. If this performs
    a mutation as part of the walk, it will store the mutated node in the
    `mutant` attribute. If the walk does not result in any mutation, `mutant`
    will be `None`.

    Note that `mutant` is just the specifically mutated node. It will generally
    be a part of the larger AST which is returned from `walk()`.
    """

    def __init__(self, occurrence, operator):
        self.operator = operator
        self._occurrence = occurrence
        self._count = 0
        self._mutation_applied = False

    @property
    def mutation_applied(self):
        "Whether this visitor has applied a mutation."
        return self._mutation_applied

    def visit(self, node):
        for index, _ in enumerate(self.operator.mutation_positions(node)):
            if self._count == self._occurrence:
                self._mutation_applied = True
                node = self.operator.mutate(node, index)
            self._count += 1

        return node


def _make_diff(original_source, mutated_source, module_path):
    module_diff = ["--- mutation diff ---"]
    for line in difflib.unified_diff(
        original_source.split("\n"),
        mutated_source.split("\n"),
        fromfile="a" + str(module_path),
        tofile="b" + str(module_path),
        lineterm="",
    ):
        module_diff.append(line)
    return module_diff



"""Support for timing the execution of functions.

This is primarily intended to support baselining, but it's got some reasonable
generic functionality.
"""

import datetime


class Timer:
    """A simple context manager for timing events.

    Generally use it like this:

    .. code-block::

        with Timer() as t:
            do_something()
        print(t.elapsed())
    """

    def __init__(self):
        self._start = None
        self.reset()

    def reset(self):
        """Set the elapsed time back to 0."""
        self._start = datetime.datetime.now()

    @property
    def elapsed(self):
        """Get the elapsed time between the last call to `reset` and now.

        Returns a `datetime.timedelta` object.
        """
        return datetime.datetime.now() - self._start

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        pass



"""A tool for launching HTTP workers and executing a session using them.

This reads the 'distributor.http.worker-urls' field of a config to see where workers are expected to be
running. For each worker, it makes a clone of the git repository that's going to be tested, optionally changing to a
directory under the root of the clone before starting the worker. It then starts the workers with the correct options to
provide the configured URLs.
"""

import asyncio
import logging
import tempfile
from pathlib import Path

import click
import git
import yarl

import cosmic_ray.config

log = logging.getLogger()


async def run(config_file, repo_url, location):
    """Start the configured workers on the own git clones.

    Args:
        config_file: The Cosmic Ray configuration file describing the distributor URLs.
        repo_url: The git repository to clone for each worker.
        location: The relative path into the cloned repository to use as the cwd for
            each worker.
    """
    config = cosmic_ray.config.load_config(config_file)
    worker_urls = config.sub("distributor", "http").get("worker-urls", ())

    worker_args = tuple(_urls_to_args(worker_urls, Path(config_file).resolve()))
    if not worker_args:
        log.warning("No valid worker URLs found in config %s", config_file)

    with tempfile.TemporaryDirectory() as root:
        for idx, _ in enumerate(worker_args):
            _create_clone(repo_url, Path(root) / str(idx))

        procs = [
            await asyncio.create_subprocess_shell(
                f"cosmic-ray --verbosity INFO http-worker {option} {value}", cwd=Path(root) / str(idx) / location
            )
            for idx, (option, value) in enumerate(worker_args)
        ]

        await asyncio.gather([await proc.communicate() for proc in procs])


@click.command(help=__doc__)
@click.argument("config_file", type=click.Path(exists=True, dir_okay=False, readable=True))
@click.argument("repo_url")
@click.option("--location", default="", help="The relative path under a repo clone at which to run the worker")
def main(config_file, repo_url, location):
    logging.basicConfig(level=logging.INFO)
    asyncio.get_event_loop().run_until_complete(run(config_file, repo_url, location))


def _create_clone(source_repo_url, destination_dir):
    url = yarl.URL(source_repo_url)
    if url.scheme == "":
        url = yarl.URL.build(scheme="file", path=str(Path(url.path).resolve()))

    log.info("Cloning %s to %s", url, destination_dir)
    git.Repo.clone_from(str(url), destination_dir, depth=1)


LOCALHOST_ADDRESSES = (
    "localhost",
    "0.0.0.0",
    "127.0.0.1",
)


def _urls_to_args(urls, config_filepath: Path):
    for url in urls:
        url = yarl.URL(url)
        if url.scheme == "":
            socket_path = config_filepath.parent / url.path
            yield ("--path", socket_path)
        elif url.scheme in ("http", "https"):
            if url.port is None:
                log.warning("HTTP(S) URL %s has no port", url)

            elif url.host.lower() not in LOCALHOST_ADDRESSES:
                log.warning("%s does not appear to be on localhost", url)

            else:
                yield ("--port", url.port)
        else:
            log.warning("The scheme of URL %s is not supported", url)


if __name__ == "__main__":
    main()



"Common implementation for operators that replace keywords."

from parso.python.tree import Keyword

from .operator import Operator

# pylint: disable=E1101


class KeywordReplacementOperator(Operator):
    """A base class for operators that replace one keyword with another
    """

    def mutation_positions(self, node):
        if isinstance(node, Keyword):
            if node.value.strip() == self.from_keyword:
                yield (node.start_pos, node.end_pos)

    def mutate(self, node, index):
        assert isinstance(node, Keyword)
        assert node.value == self.from_keyword

        node.value = self.to_keyword
        return node

    @classmethod
    def examples(cls):
        return (
            (cls.from_keyword, cls.to_keyword),
        )



"""This module contains mutation operators which replace one
comparison operator with another.
"""
from enum import Enum
import itertools

import parso.python.tree

from ..ast import is_none, is_number
from .operator import Operator
from .util import extend_name


class ComparisonOperators(Enum):
    "All comparison operators that we mutate."
    Eq = '=='
    NotEq = '!='
    Lt = '<'
    LtE = '<='
    Gt = '>'
    GtE = '>='
    Is = 'is'
    IsNot = 'is not'


def _create_operator(from_op, to_op):
    @extend_name('_{}_{}'.format(from_op.name, to_op.name))
    class ReplaceComparisonOperator(Operator):
        "An operator that replaces {} with {}".format(from_op.name, to_op.name)

        def mutation_positions(self, node):
            if node.type == 'comparison':
                # Every other child starting at 1 is a comparison operator of some sort
                for _, comparison_op in self._mutation_points(node):
                    yield (comparison_op.start_pos, comparison_op.end_pos)

        def mutate(self, node, index):
            points = list(itertools.islice(self._mutation_points(node), index, index + 1))
            assert len(points) == 1
            op_idx, _ = points[0]
            mutated_comparison_op = parso.parse(' ' + to_op.value)
            node.children[op_idx * 2 + 1] = mutated_comparison_op
            return node

        @staticmethod
        def _mutation_points(node):
            for op_idx, comparison_op in enumerate(node.children[1::2]):
                if comparison_op.get_code().strip() == from_op.value:
                    rhs = node.children[(op_idx + 1) * 2]
                    if _allowed(to_op, from_op, rhs):
                        yield op_idx, comparison_op

        @classmethod
        def examples(cls):
            return (
                ('x {} y'.format(from_op.value), 'x {} y'.format(to_op.value)),
            )

    return ReplaceComparisonOperator


# Build all of the binary replacement operators
_OPERATORS = tuple(
    _create_operator(from_op, to_op)
    for from_op, to_op
    in itertools.permutations(ComparisonOperators, 2))

# Inject the operators into the module namespace
for op_cls in _OPERATORS:
    globals()[op_cls.__name__] = op_cls


def operators():
    "Iterable of all binary operator replacement mutation operators."
    return iter(_OPERATORS)


# This determines the allowed from-to mutations when the RHS is None.
_RHS_IS_NONE_OPS = {
    ComparisonOperators.Eq: [ComparisonOperators.IsNot],
    ComparisonOperators.NotEq: [ComparisonOperators.Is],
    ComparisonOperators.Is: [ComparisonOperators.IsNot],
    ComparisonOperators.IsNot: [ComparisonOperators.Is],
}

# This determines the allowed to mutations when the RHS is a number
_RHS_IS_INTEGER_OPS = set([
    ComparisonOperators.Eq,
    ComparisonOperators.NotEq,
    ComparisonOperators.Lt,
    ComparisonOperators.LtE,
    ComparisonOperators.Gt,
    ComparisonOperators.GtE,
])


def _allowed(to_op, from_op, rhs):
    "Determine if a mutation from `from_op` to `to_op` is allowed given a particular `rhs` node."
    if is_none(rhs):
        return to_op in _RHS_IS_NONE_OPS.get(from_op, ())

    if is_number(rhs):
        return to_op in _RHS_IS_INTEGER_OPS

    return True



"""Classes for describing work and results.
"""
import dataclasses
import enum
import pathlib
from pathlib import Path
from typing import Optional, Tuple


class StrEnum(str, enum.Enum):
    "An Enum subclass with str values."


class WorkerOutcome(StrEnum):
    """Possible outcomes for a worker."""

    NORMAL = "normal"  # The worker exited normally, producing valid output
    EXCEPTION = "exception"  # The worker exited with an exception
    ABNORMAL = "abnormal"  # The worker did not exit normally or with an exception (e.g. a segfault)
    NO_TEST = "no-test"  # The worker had no test to run
    SKIPPED = "skipped"  # The job was skipped (worker was not executed)


class TestOutcome(StrEnum):
    """A enum of the possible outcomes for any mutant test run."""

    SURVIVED = "survived"
    KILLED = "killed"
    INCOMPETENT = "incompetent"


@dataclasses.dataclass(frozen=True)
class WorkResult:
    """The result of a single mutation and test run."""

    worker_outcome: WorkerOutcome
    output: Optional[str] = None
    test_outcome: Optional[TestOutcome] = None
    diff: Optional[str] = None

    def __post_init__(self):
        if self.worker_outcome is None:
            raise ValueError("Worker outcome must always have a value.")

        if self.test_outcome is not None:
            object.__setattr__(self, "test_outcome", TestOutcome(self.test_outcome))

        object.__setattr__(self, "worker_outcome", WorkerOutcome(self.worker_outcome))

    @property
    def is_killed(self):
        "Whether the mutation should be considered 'killed'"
        return self.test_outcome != TestOutcome.SURVIVED


@dataclasses.dataclass(frozen=True)
class MutationSpec:
    "Description of a single mutation."
    module_path: Path
    operator_name: str
    occurrence: int

    # pylint: disable=R0913
    def __post_init__(self):
        object.__setattr__(self, "module_path", pathlib.Path(self.module_path))
        object.__setattr__(self, "occurrence", int(self.occurrence))


# TODO: I'm suspicious of this distinction between MutationSpec and ResolvedMutationSpec. We only really need it because
# the data we send to distributors doesn't include  start_/end_pos, and I wanted to make that clear. Am I being
# hoodwinked by type annotations? Should I just stop worrying and merge them?
@dataclasses.dataclass(frozen=True)
class ResolvedMutationSpec(MutationSpec):
    "A MutationSpec with the location of the mutation resolved."
    start_pos: Tuple[int, int]
    end_pos: Tuple[int, int]

    # pylint: disable=R0913
    def __post_init__(self):
        super().__post_init__()
        if self.start_pos[0] > self.end_pos[0]:
            raise ValueError("Start line must not be after end line")

        if self.start_pos[0] == self.end_pos[0]:
            if self.start_pos[1] >= self.end_pos[1]:
                raise ValueError("End position must come after start position.")


@dataclasses.dataclass(frozen=True)
class WorkItem:
    """A collection (possibly empty) of mutations to perform for a single test.

    This ability to perform more than one mutation for a single test run is how we support
    higher-order mutations.
    """

    job_id: str
    mutations: Tuple[ResolvedMutationSpec]

    @classmethod
    def single(cls, job_id, mutation: ResolvedMutationSpec):
        """Construct a WorkItem with a single mutation.

        Args:
            job_id: The ID of the job.
            mutation: The single mutation for the WorkItem.

        Returns:
            A new `WorkItem` instance.
        """
        return cls(job_id, (mutation,))



"""Setup for Cosmic Ray.
"""

from pathlib import Path

from setuptools import setup, find_packages


INSTALL_REQUIRES = [
    "aiohttp",
    "anybadge",
    "astunparse",
    "click",
    "decorator",
    "exit_codes",
    "gitpython",
    "parso",
    "pathlib",
    "qprompt",
    "rich",
    "sqlalchemy",
    "stevedore",
    "toml",
    "virtualenv<=16.7.10",
    "yattag",
]

setup(
    name="cosmic_ray",
    version="8.3.2",
    packages=find_packages("src"),
    author="Sixty North AS",
    author_email="austin@sixty-north.com",
    description="Mutation testing",
    license="MIT License",
    keywords="testing",
    package_dir={"": "src"},
    url="http://github.com/sixty-north/cosmic-ray",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Testing",
    ],
    python_requires=">=3.5",
    platforms="any",
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        "test": ["hypothesis", "pytest", "pytest-mock"],
        "dev": ["flake8", "black", "bumpversion", "twine"],
        "docs": ["sphinx", "sphinx_rtd_theme"],
    },
    entry_points={
        "console_scripts": [
            "cosmic-ray = cosmic_ray.cli:main",
            "cr-html = cosmic_ray.tools.html:report_html",
            "cr-report = cosmic_ray.tools.report:report",
            "cr-badge = cosmic_ray.tools.badge:generate_badge",
            "cr-rate = cosmic_ray.tools.survival_rate:format_survival_rate",
            "cr-xml = cosmic_ray.tools.xml:report_xml",
            "cr-filter-operators = cosmic_ray.tools.filters.operators_filter:main",
            "cr-filter-pragma = cosmic_ray.tools.filters.pragma_no_mutate:main",
            "cr-filter-git = cosmic_ray.tools.filters.git:main",
            "cr-http-workers = cosmic_ray.tools.http_workers:main",
        ],
        "cosmic_ray.test_runners": [
            "unittest = cosmic_ray.testing.unittest_runner:UnittestRunner",
        ],
        "cosmic_ray.operator_providers": [
            "core = cosmic_ray.operators.provider:OperatorProvider",
        ],
        "cosmic_ray.distributors": [
            "http = cosmic_ray.distribution.http:HttpDistributor",
            "local = cosmic_ray.distribution.local:LocalDistributor",
        ],
    },
    long_description=Path("README.rst").read_text(encoding="utf-8"),
)



"""Implementation of the NumberReplacer operator.
"""

import parso

from ..ast import is_number
from .operator import Operator

# List of offsets that we apply to numbers in the AST. Each index into the list
# corresponds to single mutation.
OFFSETS = [
    +1,
    -1,
]


class NumberReplacer(Operator):
    """An operator that modifies numeric constants."""

    def mutation_positions(self, node):
        if is_number(node):
            for _ in OFFSETS:
                yield (node.start_pos, node.end_pos)

    def mutate(self, node, index):
        """Modify the numeric value on `node`."""

        assert index < len(OFFSETS), 'received count with no associated offset'
        assert isinstance(node, parso.python.tree.Number)

        val = eval(node.value) + OFFSETS[index]  # pylint: disable=W0123
        return parso.python.tree.Number(' ' + str(val), node.start_pos)

    @classmethod
    def examples(cls):
        return (
            ('x = 1', 'x = 2'),
            ('x = 1', 'x = 0', 1),
            ('x = 4.2', 'x = 5.2'),
            ('x = 4.2', 'x = 3.2', 1),
            ('x = 1j', 'x = (1+1j)'),
            ('x = 1j', 'x = (-1+1j)', 1),
        )



"""Implementation of the boolean replacement operators.
"""

import parso.python.tree

from .keyword_replacer import KeywordReplacementOperator
from .operator import Operator


class ReplaceTrueWithFalse(KeywordReplacementOperator):
    """An that replaces True with False."""

    from_keyword = "True"
    to_keyword = "False"


class ReplaceFalseWithTrue(KeywordReplacementOperator):
    """An that replaces False with True."""

    from_keyword = "False"
    to_keyword = "True"


class ReplaceAndWithOr(KeywordReplacementOperator):
    """An operator that swaps 'and' with 'or'."""

    from_keyword = "and"
    to_keyword = "or"

    @classmethod
    def examples(cls):
        return (("x and y", "x or y"),)


class ReplaceOrWithAnd(KeywordReplacementOperator):
    """An operator that swaps 'or' with 'and'."""

    from_keyword = "or"
    to_keyword = "and"

    @classmethod
    def examples(cls):
        return (("x or y", "x and y"),)


class AddNot(Operator):
    """
    An operator that adds the 'not' keyword to boolean expressions.

    NOTE: 'not' as unary operator is mutated in
     ``unary_operator_replacement.py``, including deletion of the same
     operator.
    """

    NODE_TYPES = (parso.python.tree.IfStmt, parso.python.tree.WhileStmt, parso.python.tree.AssertStmt)

    def mutation_positions(self, node):
        if isinstance(node, self.NODE_TYPES):
            expr = node.children[1]
            yield (expr.start_pos, expr.end_pos)
        elif isinstance(node, parso.python.tree.PythonNode) and node.type == "test":
            # ternary conditional
            expr = node.children[2]
            yield (expr.start_pos, expr.end_pos)

    def mutate(self, node, index):
        assert index == 0

        if isinstance(node, self.NODE_TYPES):
            expr_node = node.children[1]
            mutated_code = " not{}".format(expr_node.get_code())
            mutated_node = parso.parse(mutated_code)
            node.children[1] = mutated_node

        else:
            assert node.type == "test"
            expr_node = node.children[2]
            mutated_code = " not{}".format(expr_node.get_code())
            mutated_node = parso.parse(mutated_code)
            node.children[2] = mutated_node

        return node

    @classmethod
    def examples(cls):
        return (
            ("if True or False: pass", "if not True or False: pass"),
            ("A if B else C", "A if not B else C"),
            ("assert isinstance(node, ast.Break)", "assert not isinstance(node, ast.Break)"),
            ("while True: pass", "while not True: pass"),
        )
